<?php

/**
 * @file
 * Entity List Field class file.
 */

class EntityListFieldList {

  public $fieldName;
  public $hostItem;
  public $config;
  public $delta;

  /**
   * Generate the list.
   *
   * @param string $field_name
   *   The field name.
   * @param array $host_item
   *   An array describing the item to which this field is attached. See below.
   * @param array $config
   *   The config options for the list, in an associative array.  See below.
   * @param string $delta
   *   The delta.
   *
   * The $host_item array:
   *   $host_item[0] The host item's entity_id.
   *   $host_item[1] The host item's revision_id.
   *   $host_item[2] The host item's bundle.
   *   $host_item[3] The host item's entity type.
   *
   * We don't need $host_item[1] or $host_item[2] for this class to work,
   * but we have them as a result of our call to entity_extract_ids() in the
   * module file.
   *
   * The $config array:
   *   $config['type']
   *   $config['bundle']
   *   $config['view_mode']
   *   $config['numitems']
   *   $config['sort']
   *   $config['direction']
   *   $config['terms']
   *   $config['pager']
   *   $config['cache']
   *   $config['status']
   */
  public function __construct($field_name, $host_item, $config, $delta = 0) {

    $this->fieldName = $field_name;
    $this->hostItem = $host_item;
    $this->config = $config;
    $this->delta = $delta;

  }

  /**
   * Return the list as themed output.
   *
   * @param string $page
   *   The page we're on. For paged results, this will be '0' or an empty
   *   string for the first page of results, '1' for the second page, and so on.
   *
   * @return string
   *   The themed output.
   */
  public function getList($page = '') {

    /*
     * $this->config contains the end user's "settings" for this list,
     * e.g. what the user defines in the entity's create/edit forms.
     */
    $config = $this->config;

    if ('0' === $config['numitems']) {
      return '';
    }

    $type = isset($config['type']) ? $config['type'] : 'node';
    $bundle = isset($config['bundle']) ? $config['bundle'] : json_encode(array('article'));
    $view_mode = isset($config['view_mode']) ? $config['view_mode'] : 'teaser';
    $numitems = isset($config['numitems']) ? $config['numitems'] : '3';
    $sort = isset($config['sort']) ? $config['sort'] : 'created';
    $direction = isset($config['direction']) ? $config['direction'] : 'DESC';
    $terms = isset($config['terms']) ? $config['terms'] : '';
    $pager = isset($config['pager']) ? $config['pager'] : '2';
    $cache = isset($config['cache']) ? $config['cache'] : '1';
    // Status is not implemented as something that can be configured.
    $status = isset($config['status']) ? $config['status'] : 1;

    $bundles = json_decode($bundle);

    // Check if the user can view the items.
    $access = $this->accessCheck($type, $bundles);
    if (!$access[0]) {
      return '';
    }
    elseif ('node' == $type) {
      // Bundles the user can view.
      $bundles = $access[1];
    }

    // Always require one or more bundles.
    if (empty($bundles)) {
      return '';
    }

    if ('1' === $cache) {
      // Set a cache ID based on the config.
      $cid['type'] = $type;
      $cid['bundle'] = implode(',', $bundles);
      $cid['view_mode'] = $view_mode;
      $cid['numitems'] = $numitems;
      $cid['sort'] = $sort;
      $cid['direction'] = $direction;
      $cid['terms'] = $terms;
      $cid['status'] = $status;
      // We also need the page for uniqueness.
      $cid['page'] = $page;
      // Add the language.
      global $language;
      $lang = $language->language;
      $cid['lang'] = $lang;
      $cid = 'entity_list_field:' . implode(':', $cid);
      // Return the cached data if we have any.
      if ($cachedata = cache_get($cid)) {
        if ('2' === $pager) {
          $this->entityListFieldPager();
        }
        return $cachedata->data;
      }
    }

    /* Tell EntityFieldQuery and the entity's view function what page
     * we're on. We can set this any time before we execute the query.
     */
    $_GET['page'] = $page;

    $query = new EntityFieldQuery();
    $query->entityCondition('entity_type', $type);

    // Don't display the host entity in its own list.
    if ($type == $this->hostItem[3]) {
      if (!isset($bundles) || in_array($this->hostItem[2], $bundles)) {
        $query->entityCondition('entity_id', $this->hostItem[0], '!=');
      }
    }

    $prop_info = entity_get_property_info($type);
    $props = $prop_info['properties'];

    // Deal with lingering inconsistencies re: how entities store bundle data.
    // Checking the type property works seemingly for everything except
    // taxonomy_term bundles.
    if (isset($props['type'])) {
      $query->propertyCondition('type', $bundles);
    }
    // This entityCondition works in some cases, taxonomy_term being one
    // of those cases.
    elseif ('taxonomy_term' == $type) {
      $query->entityCondition('bundle', $bundles);
    }

    // Not all entity types have a status property.
    if (isset($props['status'])) {
      $query->propertyCondition('status', $status);
    }

    // Not all entity types have a sort property.
    if (isset($props[$sort])) {
      $query->propertyOrderBy($sort, $direction);
    }

    /* @see field_info_instances()
     * field_info_instances() returns data about the bundle field
     * instances. See if any bundles have a taxonomy field and if so,
     * what is the fieldname.  We use $bundle_fields here and in the call
     * to safeLoop().
     */
    $bundle_fields = array();
    foreach ($bundles as $bundle) {
      $bundle_fields += field_info_instances($type, $bundle);
    }

    $terms = trim($terms);
    if (!empty($terms)) {
      $terms = explode(',', $terms);
      foreach ($bundle_fields as $tname => $value) {
        if ('taxonomy' == $value['widget']['module']) {
          /* The $tname is the taxonomy field name.
           * This works like an OR query.  If the item has been assigned
           * to any of the specified terms, it passes the condition.
           */
          $query->fieldCondition($tname, 'tid', $terms);
          /* We break after we find one term reference field in any of the
           * bundles. Thus, this only supports cases where we're ok with
           * ignoring any other term reference fields besides the first one
           * in $bundle_fields.  We expect all bundles to have whatever is the
           * first term reference field we find, but no errors result if a
           * bundle doesn't have a term reference field.
           */
          break;
        }
      }
    }

    if ('0' !== $pager) {
      $query->pager($numitems);
    }
    else {
      $query->range(0, $numitems);
    }

    $result = $query->execute();

    if (!$result || !isset($result[$type])) {
      return '';
    }
    else {
      $ids = array_keys($result[$type]);
      $entities = entity_load($type, $ids);
    }

    if ($entities) {
      $entities = $this->safeLoop($entities, $this->config, $bundle_fields);
      $output = $this->theme($entities, $this->config, $page);
    }
    else {
      $output = '';
    }

    if ('1' === $cache) {
      cache_set($cid, $output, 'cache', CACHE_TEMPORARY);
    }

    return $output;

  }

  /**
   * Theme the list.
   *
   * @param array $entities
   *   The entities to list.
   * @param string $config
   *   The list config as defined in the widget form.
   *
   * @return string
   *   The themed output.
   */
  protected function theme($entities, $config) {

    extract($config);

    if ('2' === $pager) {
      $this->entityListFieldPager();
    }

    $items = $this->renderItems($entities, $type, $view_mode);

    $output = theme('item_list', array('items' => $items));
    if ('0' !== $pager) {
      $output .= theme('pager');
    }

    return $output;
  }

  /**
   * Render list items.
   *
   * @param array $entities
   *   The entities to list.
   * @param string $type
   *   The entity type.
   * @param string $view_mode
   *   The view mode.
   *
   * @return array
   *   The rendered items.
   */
  protected function renderItems($entities, $type, $view_mode) {

    $viewfunc = $type . '_view';

    if (function_exists($viewfunc)) {
      foreach ($entities as $entity) {
        $items[] = drupal_render($viewfunc($entity, $view_mode));
      }
    }
    elseif (function_exists('theme_entity_list_field_' . $type)) {
      // How to provide a custom theme function in cases when an entity type
      // has no view function, as is the case with the taxonomy_vocabulary type.
      // See theme_entity_list_field_taxonomy_vocabulary() for an example.
      $vars['entities'] = $entities;
      $items = theme('entity_list_field_' . $type, $vars);
    }
    else {
      $msg = 'No view function or theme function in EntityListFieldList::renderItems().';
      watchdog('entity_list_field', $msg);
      // No view function and no theme function, so just list the titles.
      foreach ($entities as $entity) {
        $items[] = $entity->name;
      }
    }

    return $items;
  }

  /**
   * Activate the ajax pager.
   */
  protected function entityListFieldPager() {

    $path = drupal_get_path('module', 'entity_list_field');
    drupal_add_js($path . '/ajax-pager.js');

    $js_settings = array(
      'field_name' => $this->fieldName,
      'host_item' => $this->hostItem,
      'item' => $this->config,
      'delta' => $this->delta,
    );
    drupal_add_js(array('entity_list_field' => $js_settings), 'setting');
  }

  /**
   * Check if the user can view the list items.
   *
   * @param array $type
   *   The entity type.
   * @param string $bundles
   *   The bundles.
   *
   * @return array
   *   Info about the user's access to this type and its bundles.
   *   $info[0] bool -- Can we view the type?
   *   $info[1] array -- For nodes, the bundles we can view.
   */
  protected function accessCheck($type, $bundles) {

    $info = array(TRUE, array());

    if ('node' == $type) {

      if (!user_access('access content')) {
        $info[0] = FALSE;
      }
      else {
        // Add the node_view_permissions integration.
        $bundles = $this->nodeViewPermissions($bundles);
        $info[1] = $bundles;
      }

    }
    elseif ('user' == $type) {
      if (!user_access('access user profiles')) {
        $info[0] = FALSE;
      };
    }
    elseif ('comment' == $type) {
      if (!user_access('access comments')) {
        $info[0] = FALSE;
      };
    }
    return $info;
  }

  /**
   * Add the node_view_permissions integration.
   *
   * @param array $bundles
   *   Bundles to check.
   *
   * @return array
   *   Bundles the user is allowed to view.
   */
  protected function nodeViewPermissions($bundles) {
    if (module_exists('node_view_permissions')) {
      $nvp_bundles = node_view_permissions_get_configured_types();
      foreach ($nvp_bundles as $key => $bundle) {
        $perm = 'view any ' . $bundle . ' content';
        if (!user_access($perm)) {
          $key = array_search($bundle, $bundles);
          unset($bundles[$key]);
        };
      }
    }
    return $bundles;
  }

  /**
   * Prevent infinite loops.
   *
   * Prevent display of Entity List fields in list items.
   *
   * @param array $entities
   *   The entities in the list.
   * @param array $config
   *   The list config as defined in the widget form.
   * @param array $bundle_fields
   *   Info about a bundle's field instances, as furnished
   *   by field_info_instances().
   *
   * @return array
   *   Entities.
   */
  protected function safeLoop($entities, $config, $bundle_fields) {

    $has = FALSE;

    extract($config);

    foreach ($bundle_fields as $value) {
      // Check if any of the bundles have an Entity List field.
      if ('entity_list_field' == $value['widget']['module']) {
        $fn = $value['field_name'];
        $has[$fn] = $fn;
      }
    }

    if ($has) {
      foreach ($has as $name) {
        // We need to remove the list fields from the items.
        foreach ($entities as $key => $entity) {
          if (isset($entity->{$name})) {
            unset($entity->{$name});
          }
          $entities[$key] = $entity;
        }
      }
    }

    return $entities;

  }

}
